;;; sdbm egg

;; Todo: Determine if fetch should throw an error on missing key (instead of returning #f)

;;; Changes over sdbm release

;; Page and dir block size configurable at database open time.
;; Byte-order configurable at compile time
;;   (-Dsdbm-little-endian, -Dsdbm-big-endian; default: host order).
;; Hash configurable at compile time
;;   (-Dsdbm-hash-djb: DJB hash; default: native SDBM hash).

;;; Module

(module sdbm
(open-database close-database fetch store! delete!
 pair-iterator next-pair
 pair-fold)

(import scheme chicken)
(require-library posix lolevel data-structures srfi-4)
(import posix
        (only lolevel move-memory!)
        (only data-structures ->string)
        srfi-4)

;;; hash functions

(define (djb-hash str len)
  ;; Warning: (fxshl #x3fffffff 1) = #xffffffff on 32-bit systems.
  (let loop ((hash 5381) (i 0))
    (if (fx>= i len)
        hash
        (let ((c (char->integer (string-ref str i))))   ; if len<strlen, can use *byte-string-ref
          (loop (fx+ c (fx+ hash (fxshl hash 5)))
                (fx+ i 1))))))

(define (native-hash str len) ;; /* n = *str++ + 65599 * n; */
  (let loop ((hash 0) (i 0))
    (if (fx>= i len)
        hash
        (let ((c (char->integer (string-ref str i))))
          (loop (fx- (fx+ c (fx+ (fxshl hash 6) (fxshl hash 16)))
                     hash)
                (fx+ i 1))))))

(define (make-byte-string x) (make-string x))
(define (byte-string-length x) (string-length x))
(define (byte-string-ref x i) (string-ref x i))

;;; Local blob/u8vector handling

(include "blobbie.scm")

;; Configure this egg at compile-time for little-endian, big-endian or
;; host-order data access.

(cond-expand
 (sdbm-little-endian
  (define (blob-u16-ref b i) (blob-le16-ref b i))
  (define (blob-u16-set! b i n) (blob-le16-set! b i n))
  (define (u8vector-u16-ref v i) (u8vector-le16-ref v i))
  (define (*u8vector-u16-ref v i) (*u8vector-le16-ref v i))
  (define (u8vector-u16-set! v i n) (u8vector-le16-set! v i n))
  (define (*u8vector-u16-set! v i n) (*u8vector-le16-set! v i n)))
 (sdbm-big-endian
  (define (blob-u16-ref b i) (blob-be16-ref b i))
  (define (blob-u16-set! b i n) (blob-be16-set! b i n))
  (define (u8vector-u16-ref v i) (u8vector-be16-ref v i))
  (define (*u8vector-u16-ref v i) (*u8vector-be16-ref v i))
  (define (u8vector-u16-set! v i n) (u8vector-be16-set! v i n))
  (define (*u8vector-u16-set! v i n) (*u8vector-be16-set! v i n)))
 (else
  (define (blob-u16-ref b i) (blob-hu16-ref b i))
  (define (blob-u16-set! b i n) (blob-hu16-set! b i n))
  (define (u8vector-u16-ref v i) (u8vector-hu16-ref v i))
  (define (*u8vector-u16-ref v i) (*u8vector-hu16-ref v i))
  (define (u8vector-u16-set! v i n) (u8vector-hu16-set! v i n))
  (define (*u8vector-u16-set! v i n) (*u8vector-hu16-set! v i n))))

;;; debugging

(cond-expand (debug
              (define (dprintf . args) (apply fprintf (current-error-port) args)))
             (else
              (define-syntax dprintf (syntax-rules () ((dprintf . args) (begin))))))

;;; main

(define-constant +page-extension+ ".pag")
(define-constant +dir-extension+ ".dir")
(define-constant +max-splits+ 10)
(define-constant +page-block-power+ 12)
(define-constant +dir-block-power+ 12)
(define (sdbm-hash-partial key len)
  (cond-expand (sdbm-hash-djb
                (djb-hash key len))
               (else
                (native-hash key len))))
(define (sdbm-hash key)
  (sdbm-hash-partial key (string-length key)))

(define-record sdbm
  page-fd dir-fd             ; File descriptors
  page-buf dir-buf           ; Blob representing current page/dir block in memory
  page-block dir-block       ; Block address of loaded page/dir block, or -1
  hash-bit dir-bit           ; Last hash and dir bit #, obtained during page read 
  dir-bits                   ; Number of bits in directory
  )

(define (sdbm-page-block-size db)     ; block size (in bytes!)
  (blob-size (sdbm-page-buf db)))
(define (sdbm-dir-block-size db)
  (blob-size (sdbm-dir-buf db)))

;; flags, mode as in e.g. Unit posix file-open
;; page-block-power: size of each data page block, as a power of 2
;; dir-block-power: size of each directory block, as a power of 2
(define (open-database name #!key
                       (flags (+ open/rdwr open/creat open/binary))
                       (mode (+ perm/irwxu perm/irgrp perm/iroth))
                       (page-block-power +page-block-power+)
                       (dir-block-power +dir-block-power+))
  (let ((page (file-open (string-append name +page-extension+) flags mode))
        (dir  (file-open (string-append name +dir-extension+)  flags mode)))
    (let ((dir-size (file-size dir)))
      (make-sdbm page dir
                 (make-zero-blob (arithmetic-shift 1 page-block-power)) ; page-buf
                 (make-zero-blob (arithmetic-shift 1 dir-block-power)) ; dir-buf
                 -1                                                            ; page-block
                 (if (= dir-size 0) 0 -1)    ; dir-block (-1 if we need to load it)
                 0 0                         ; hash-bit, dir-bit
                 (* dir-size 8)              ; dir-bits
                 ))))

(define (close-database db)
  (file-close (sdbm-page-fd db))
  (file-close (sdbm-dir-fd db)))

(define (key->string key)
  (cond ((string? key) key)
        ((symbol? key) (##sys#symbol->string key))
        (else (->string key))))

(define (fetch db key)
  (let ((key (key->string key)))
    (read-page! db (sdbm-hash key))
    (get-pair (sdbm-page-buf db) key)))

(define (read-page! db hash)
  (read-page-block! db (calc-page-block db hash)))
(define (calc-page-block db hash)
  (bitwise-and hash (calc-hash-mask db hash)))
(define (write-page! db)
  (write-page-block! db (sdbm-page-buf db) (sdbm-page-block db)))

(define (left-child dbit)       ; dbit's left child (as bit offset)
  (fx+ (fx* 2 dbit) 1))
(define (right-child dbit)
  (fx+ (fx* 2 dbit) 2))
(define (left-child? hash hbitmask)    ; Is this hash a left or right child, looking at hbit?
  (fx= 0 (bitwise-and hash hbitmask)))

;; WARNING: side-effects: sets directory and hash bit in db record.
;; This is underhanded.
(define (calc-hash-mask db hash)
  (let loop ((dbit 0)
             (hbit 0))
    (let ((hbit-mask (fxshl 1 hbit)))
      (cond ((internal-node? db dbit)
             (loop (if (fx= 0 (bitwise-and hash hbit-mask))
                       (left-child dbit)
                       (right-child dbit))
                   (fx+ hbit 1)))
            (else
             (sdbm-dir-bit-set! db dbit)
             (sdbm-hash-bit-set! db hbit)
             (fx- hbit-mask 1)     ; hash-mask
             )))))

;; The database directory is a radix 2-trie.  The file format
;; represents each node as a single bit at offset N, where 0 means the
;; node is a leaf (no children) and 1 means the node has children,
;; which are located at offsets 2N+1 (left) and 2N+2 (right).

;; To look up the bucket address of a hash key, start at bit d=0 of the
;; directory and bit h=0 of the hash.  If directory bit d is one,
;; examine bit h of the hash, and go to d's left child if zero, or the
;; right child if one, increasing h by 1.  Repeat until directory bit d
;; is zero, then take the lower h bits of the hash as the page address.
;; (Looking at 0 bits results in an address of 0.) 

;; Find bit number BIT in database directory and return #t if set,
;; #f if not.  BIT can be out of range, in which case this returns #f.
(define (dir-bit db bit)   ; #t if internal node, #f if leaf node
  (if (>= bit (sdbm-dir-bits db))
      #f
      (let ((byte (arithmetic-shift bit -3))
            (bit (bitwise-and bit 7)))
        (let ((block (fx/ byte (sdbm-dir-block-size db))) ; NB   fxshr works here
              (byte (fxmod byte (sdbm-dir-block-size db)))) ; NB   fxand works here
          (read-dir-block! db block)
          (extract-bit (sdbm-dir-buf db) byte bit)))))
(define internal-node? dir-bit)
(define (set-dir-bit! db dbit)
  (let ((byte (arithmetic-shift dbit -3))               ;; FIXME dup with dir-bit
        (bit (bitwise-and dbit 7)))
    (let ((block (fx/ byte (sdbm-dir-block-size db)))
          (byte (fxmod byte (sdbm-dir-block-size db))))
      (read-dir-block! db block)
      (set-bit! (sdbm-dir-buf db) byte bit)
      (write-dir-block! db)
      (when (>= dbit (sdbm-dir-bits db)) ; Update number of dir bits.
        (sdbm-dir-bits-set! db (fx+ 1 dbit))))))        ;; FIXME Should increase to page boundary
                                                        ;; FIXME anyway, is fx+ 1 correct?

(define (extract-bit blob byte bit)   ; #t if bit set, #f if not set
  (let ((val (blob-u8-ref blob byte)))
    (not (fx= 0 (bitwise-and val (fxshl 1 bit))))))
(define (set-bit! blob byte bit)      ; meh, we should accept val
  (let ((val (blob-u8-ref blob byte)))
    (blob-u8-set! blob byte
                  (bitwise-ior val (fxshl 1 bit)))))

;; * Special note: The original code does not zero the directory block
;;   before reading in; I can't see how this works in all cases.
;; * Although we zero the entire dir block, we could and should just zero
;;   what we can't read.
(define (read-dir-block! db block)     ;; Make BLOCK the current dir block.  Return unspecified.
  (cond ((= (sdbm-dir-block db) block)
         (void))
        (else
         (let* ((fd (sdbm-dir-fd db))
                (buf (sdbm-dir-buf db))
                (size (blob-size buf)))
           (dprintf "Reading directory block ~a\n" block)
           (set-file-position! fd (* block size))                 ; fxshl
           (blob-zero! buf)
           (file-read fd size (sdbm-dir-buf db))
           (sdbm-dir-block-set! db block)))))

(define (write-dir-block! db)           ;; Write current dir block out.  Return unspecified.
  (let ((block (sdbm-dir-block db))
        (size (sdbm-dir-block-size db))
        (fd (sdbm-dir-fd db)))
    (dprintf "Writing directory block ~a\n" block)
    (set-file-position! fd (* block size))
    (file-write fd (sdbm-dir-buf db))
    (void)))

;; Read block# into current page buffer.
;; Returns: #t on success, #f if at EOF, error on short read.
(define (read-page-block! db block)
  (cond ((= (sdbm-page-block db) block)
         #t)
        (else
         (let ((size (sdbm-page-block-size db))
               (buf (sdbm-page-buf db)))
           (dprintf "Reading page block ~a\n" block)
           (set-file-position! (sdbm-page-fd db)
                               (* block size))                 ; fxshl
           (let ((sz (cadr
                      (file-read (sdbm-page-fd db) size buf))))
             ;; NB: If caller always tests return value, we could leave the
             ;; old block cached.
             (sdbm-page-block-set! db block)
             (cond ((= sz 0)
                    (blob-zero! buf)
                    #f)
                   ((= sz size)
                    #t)
                   (else
                    (error 'read-page-block! "short read at block" block))))))))

(define (write-page-block! db page block) ; write buffer PAGE into block #
  (let ((size (blob-size page))
        (fd (sdbm-page-fd db)))
    (dprintf "Writing page block ~a\n" block)
    (set-file-position! fd (* block size))
    (file-write fd page)))

;;; Page format

;; page format:
;;      +------------------------------+
;; ino  | n | keyoff | valoff | keyoff |  header: all u16 (network byte order)
;;      +------------+--------+--------+  n: # of entries (keys + datas)
;;      | valoff | - - - ---->         |
;;      +--------+---------------------+
;;      |        F R E E A R E A       |
;;      +--------------+---------------+
;;      |  <---- - - - | data          |
;;      +--------+-----+----+----------+
;;      |  key   | data     | key      |
;;      +--------+----------+----------+
;;
;; calculating the offsets for free area:  if the number
;; of entries (ino[0]) is zero, the offset to the END of
;; the free area is the block size. Otherwise, it is the
;; nth (ino[ino[0]]) entry's offset.

;;; Items (key-value pairs)

;; Common named-loop construct over items in page.  The inner loop
;; demands a macro implementation for speed and also we can implement
;; for-each, map, fold, and find-first semantics over this.
;; (foof-loop / fast-loop with a custom page iterator is another
;;  possible choice.)

(define-syntax page-loop
  (syntax-rules (=>)
    ((_ (page-blob)                              ; page-loop args
        (u8buf keyoff keylen valoff vallen i)    ; body arg identifiers
        loop-name ((lvar lval) ...)              ; loop name/args
        => res                                   ; loop result (loop args visible)
        body ...)                                ; loop body
     (let* ((buf page-blob)
            (n (page-items buf))
            (size (blob-size buf)))
       (if (>= (fxshl (fx+ n 1) 1) size)
           (error 'page-fold "number of page items out of range" n)
           (let ((u8buf (blob->u8vector/shared buf)))
             (let loop ((i 1) (end size)
                        (lvar lval) ...)
               (if (> i n)
                   res
                   (let* ((keyoff (*u8vector-u16-ref u8buf (fxshl i 1)))
                          (keylen (fx- end keyoff))
                          (valoff (*u8vector-u16-ref u8buf (fxshl (fx+ i 1) 1)))
                          (vallen (fx- keyoff valoff)))
                     (letrec-syntax ((loop-name (syntax-rules ()
                                                  ((_ . args)
                                                   (loop (fx+ i 2) valoff
                                                         . args)))))
                       body ...))))))))))

;; Page utility functions

(define (page-fold page kons knil)
  (page-loop (page) (u8buf koff klen voff vlen i)
             loop ((kvs knil)) => kvs
             (let ((key (blob-substring page koff klen))
                   (val (blob-substring page voff vlen)))
               (loop (kons key val kvs)))))
(define (page-map proc page)
  (reverse (page-fold page (lambda (k v kvs)
                             (cons (proc k v) kvs))
                      '())))
(define (page-for-each proc page)   ; proc: (lambda (key val) ...)
  (page-fold page (lambda (k v kvs) (proc k v))
             (void)))

(define (get-pair buf k)
  (find-key buf k #f))
(define (pair-exists? buf k)
  (find-key buf k #t))

(define (find-key buf key exist-only?)
  (let ((keylen (byte-string-length key)))
    (page-loop (buf) (u8buf koff klen voff vlen i)
               loop () => #f
               (if (fx= keylen klen)
                   (if (u8vector-substring=? u8buf key koff klen)
                       (or exist-only?
                           (blob-substring buf voff vlen))
                       (loop))
                   (loop)))))

(define (page-items page)         ; Return number of u16 header items (key or data offsets)
  (blob-u16-ref page 0))
(define (set-page-items! page n)
  (blob-u16-set! page 0 n))

(define (bytes-free-in-page page)  ; Return bytes free in page (including header bytes)
  (let* ((n (page-items page))
         (nb (fxshl n 1)))
    (let ((free-start (fx+ 2 nb))
          (free-end (if (fx= 0 nb)
                        (blob-size page)
                        (blob-u16-ref page nb))))
      (fx- free-end free-start))))

(define (pair-fits? page key val)
  (<
   (fx+ (byte-string-length key)
        (byte-string-length val))
   (fx- (bytes-free-in-page page) (fx+ 2 2)))) ; header bytes for keyoff, valoff

(define (pair-too-big? page key val)           ; really this is property of db not page
  (> (fx+ (byte-string-length key)
          (byte-string-length val))
      (fx- (blob-size page) 6)))      ; 6 bytes for u16 item count, key offset, data offset

;; Assumes there are enough bytes in the page.  Call pair-fits? first to be sure.
(define (put-pair! page key val)
  (let* ((n (page-items page))
         (nb (fxshl n 1))
         (free-end (if (fx= 0 nb)
                       (blob-size page)
                       (blob-u16-ref page nb))))
    (let* ((keyoff (fx- free-end (string-length key)))
           (valoff (fx- keyoff (string-length val))))
      (blob-set-substring! page keyoff key)
      (blob-set-substring! page valoff val)
      (blob-u16-set! page (fx+ nb 2) keyoff)    ;; FIXME: Don't think this is right.
      (blob-u16-set! page (fx+ nb 4) valoff)
      (set-page-items! page (fx+ n 2)))))

(define (split-page! left right hbit-mask)
  ;; assumptions: RIGHT is zeroed out
  (let* ((buf (make-blob (blob-size left))))
    (move-memory! left buf)
    (blob-zero! left)
    (page-for-each
     (lambda (key val)
       (put-pair! (if (left-child? (sdbm-hash key)
                                   hbit-mask)
                      left right)
                  key val))
     buf)))

;;;

(define (store! db key val #!optional (replace #t))
  (let* ((key (key->string key))
         (hash (sdbm-hash key)))
    (read-page! db hash)
    (let ((page (sdbm-page-buf db)))
      (when (pair-too-big? page key val)
             (error 'store "pair too big for database" key val))
      (if replace
          (delete-pair! page key)              ; If make-room fails, we lose this pair
          (if (pair-exists? page key)
              (error 'store "pair already exists" key)))
      (cond ((pair-fits? page key val)
             (put-pair! page key val))
            ((make-room! db hash key val)
             (put-pair! (sdbm-page-buf db) key val)) ; make-room may change current page block
            (else
             (error 'store "unable to fit pair" key val)))
      (write-page! db)
      (void))))

;; To split a page, we need the next significant hbit so we may hash to left or right page
;; based on the new bit in the key.  We also need the current dbit so it may be set to
;; indicate a split.  Finally, we need the block address of the new page which can
;; be calculated using (hash & hmask) | hbit.  Some mathematical relations:
;; - DBIT specifies one node in the tree numbered from left to right.  DBIT children
;;   are at 2*DBIT+1 and 2*DBIT+2.
;; - DBIT's parent is at (DBIT-1)>>1.  I.e. floor((dbit-1)/2).
;; - HBIT is the node level in the directory tree, starting from 0.
;; - HBITMASK is 1 << HBIT.   I.e. b100 for tree level 2.
;; - HMASK is HBITMASK-1.
;; - HBIT is also floor(log(DBIT+1)), i.e. find the
;;   topmost bit # of DBIT+1.  The only way I know of to do this is counting divisions
;;   by 2, i.e. right shift by 1 until 0.
;; - DBIT can be calculated from HBIT and traversing the directory tree using the page address
;;   (hash) as normal--but I don't know a faster way to do it.
;; - Splitting a page with address PAGE results in two pages with
;;   addresses PAGE and PAGE|HBITMASK == PAGE|(1<<HBIT).

(define (make-room! db hash key val)    ; HASH--hash of key
  (let loop ((n +max-splits+))
    (if (fx= n 0)
        #f
        (let* ((hbit (sdbm-hash-bit db))
               (hbitmask (fxshl 1 hbit))
               (dbit (sdbm-dir-bit db)))
          (let* ((left (sdbm-page-buf db))
                 (right (make-zero-blob (blob-size left))))
            (split-page! left right hbitmask) ; Split left page pairs into right page.
            (set-dir-bit! db (sdbm-dir-bit db)) ; Reflect split in directory.
            (let* ((left-block (sdbm-page-block db))
                   (right-block (bitwise-ior left-block hbitmask)))
              (dprintf "Split page ~a to ~a\n" left-block right-block)
              ;; If this key is to be sent to the right (new) page, write the left
              ;; page and make the right page current.  Otherwise, write the right page.
              ;; The current page need not be written yet; we expect it to be
              ;; written by put-pair! when the key is stored.
              (let ((left-child? (left-child? hash hbitmask)))
                (cond (left-child?
                       (write-page-block! db right right-block))
                      (else
                       (write-page-block! db left left-block)
                       (sdbm-page-buf-set! db right)
                       (sdbm-page-block-set! db right-block)))
                (cond ((pair-fits? (sdbm-page-buf db) key val) ; Yuck.
                       #t)
                      (else
                       ;; put these as loop vars -- local update b/c no read-page!
                       (sdbm-hash-bit-set! db (fx+ hbit 1))
                       (sdbm-dir-bit-set! db (if left-child?
                                                 (left-child dbit)
                                                 (right-child dbit)))
                       ;; Going to split again, so write out deferred page.
                       (write-page! db)
                       (loop (fx- n 1)))))))))))

;; Delete pair matching KEY from PAGE (or no-op, if KEY does not
;; exist).  Deleted key/value data leaves a gap; shift data to the
;; right to close it.  Then rewrite the affected item offsets
;; (all higher-numbered items) correspondingly. Decrease # page items.
;; Returns: #t if key found, #f if not.
(define (delete-pair! page key)
  (let ((keylen (byte-string-length key))
        (n (page-items page)))
    (page-loop (page) (u8buf koff klen voff vlen i)
               loop ((shift 0)) => (if (fx= shift 0) #f #t)
               (cond
                ((fx= shift 0)
                 (if (fx= keylen klen)
                     (cond ((u8vector-substring=? u8buf key koff klen)
                            (let* ((shift (fx+ klen vlen))
                                   (free-end (*u8vector-u16-ref u8buf (fxshl n 1)))
                                   (bytes (fx- voff free-end)))
                              (dprintf "key found at index ~a" i)
                              (dprintf "shifting ~a+~a +~a\n" free-end bytes shift)
                              (move-memory! u8buf u8buf bytes
                                            free-end (fx+ free-end shift))
                              (set-page-items! page (fx- n 2))
                              (loop shift))) ; Shift remaining offsets.
                           (else (loop shift)))
                     (loop shift)))
                (else
                 ;; 
                 (dprintf "rewrite ~a koff ~a+~a\n" (fx- i 2) koff shift)
                 (dprintf "rewrite ~a voff ~a+~a\n" (fx- i 1) voff shift)
                 ;; We need to write to the PREVIOUS key/val.
                 (*u8vector-u16-set! u8buf (fxshl (fx- i 2) 1)
                                     (fx+ koff shift))
                 (*u8vector-u16-set! u8buf (fxshl (fx- i 1) 1)
                                     (fx+ voff shift))
                 (loop shift))))))

;; Delete KEY from DB or throw an error if KEY does not exist.  Note
;; that your file will never get smaller; sdbm has no provision for
;; collapsing the hash table.
(define (delete! db key)
  (let ((key (key->string key)))
    (read-page! db (sdbm-hash key))
    (unless (delete-pair! (sdbm-page-buf db) key)
      (error 'delete! "no such key" key))
    (write-page! db)
    (void)))

;;; Key folding

(define (pair-fold db kons knil)       ; kons := (lambda (k v kvs) ...) -> kvs'
  (let loop ((knil knil) (block 0))
    (if (not (read-page-block! db block))
        knil
        (let ((page (sdbm-page-buf db)))
          (let ((seed (page-fold page kons knil)))
            (loop seed (+ block 1)))))))

;;; Key iterator (firstkey/nextkey)

;; A cursor version of page-loop.  Could be
;; more elegantly implemented via page-loop + call/cc,
;; a lazy list or just drop and make FOLD the only interface.
;; Much code from page-loop is duplicated here.

;; Despite the high overhead imposed by the iterator on every
;; call to next-pair, this is only 3x slower than
;; the C implementation at retrieving all pairs.
;; However, pair-fold is still faster.

(define-record sdbm-iterator db page-block item end)

(define (get-nth-pair iter page i)
  (let ((n (page-items page))
        (size (blob-size page))
        (end (sdbm-iterator-end iter)))
    (if (>= (fxshl (fx+ n 1) 1) size)
        (error 'get-nth-pair "number of page items out of range" n)
        (and (<= i n)
             (let* ((koff (blob-u16-ref page (fxshl i 1)))
                    (klen (fx- end koff))
                    (voff (blob-u16-ref page (fxshl (fx+ i 1) 1)))
                    (vlen (fx- koff voff)))
               (dprintf "koff ~a klen ~a voff ~a vlen ~a\n" koff klen voff vlen)
               (let ((key (blob-substring page koff klen))
                     (val (blob-substring page voff vlen)))
                 (sdbm-iterator-end-set! iter voff)
                 (cons key val)))))))

;; Make a new pair iterator on database DB.
(define (pair-iterator db)
  (make-sdbm-iterator db 0 1 (sdbm-page-block-size db)))
;; Given a pair-iterator ITER, return the next pair if available, or #f
;; if there are no more pairs.
(define (next-pair iter)
  (let ((block (sdbm-iterator-page-block iter))
        (i (sdbm-iterator-item iter))
        (db (sdbm-iterator-db iter)))
    (if (read-page-block! db block)
        (let ((page (sdbm-page-buf db)))
          (let ((item (get-nth-pair iter page i)))
            (cond (item
                   (sdbm-iterator-item-set! iter (+ i 2))
                   item)
                  (else
                   (let ((next-page (+ block 1)))
                     (sdbm-iterator-page-block-set! iter next-page)
                     (sdbm-iterator-item-set! iter 1)
                     (sdbm-iterator-end-set! iter (blob-size page))
                     (next-pair iter))))))
        #f)))

) ;module
